/*
 * Copyright (c) 2009 by Red Hat Inc and/or its affiliates or by
 * third-party contributors as indicated by either @author tags or express
 * copyright attribution statements applied by the authors.  All
 * third-party contributions are distributed under license by Red Hat Inc.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution; if not, write to:
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301  USA\
 */
package org.hibernate.engine.jdbc;

import java.sql.Connection;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * TODO : javadoc
 *
 * @author Steve Ebersole
 */
public class JdbcSupportLoader {
	private static final Logger log = LoggerFactory.getLogger( JdbcSupportLoader.class );
	public static final String JDBC3_SUPPORT_CLASS = "org.hibernate.engine.jdbc.jdbc3.JdbcSupportImpl";
	public static final String JDBC4_SUPPORT_CLASS = "org.hibernate.engine.jdbc.jdbc4.JdbcSupportImpl";
	public static final Class[] JDBC4_SUPPORT_ARGS = new Class[] { boolean.class };

	public static JdbcSupport loadJdbcSupport(Connection jdbcConnection) {
		if ( isJdbc4() ) {
			final boolean lobCreationMethodsImplemented =
					jdbcConnection != null && areLobCreationMethodsImplemented( jdbcConnection );
			log.info( "loading JDBC 4 support [lobCreationMethodsImplemented=" + lobCreationMethodsImplemented + "]" );
			try {
				Object[] args = new Object[] { Boolean.valueOf( lobCreationMethodsImplemented ) };
				Class supportClass = loadClass( JDBC4_SUPPORT_CLASS );
				return ( JdbcSupport ) supportClass.getConstructor( JDBC4_SUPPORT_ARGS )
						.newInstance( args );
			}
			catch ( InvocationTargetException e ) {
				e.printStackTrace();
			}
			catch ( ClassNotFoundException e ) {
				log.warn( "Could not locate JDBC support class " + JDBC4_SUPPORT_CLASS, e );
			}
			catch ( NoSuchMethodException e ) {
				log.warn( "Could not locate appropriate JDBC support class constructor", e );
			}
			catch ( IllegalAccessException e ) {
				log.warn( "Could not instantiate JDBC support class " + JDBC4_SUPPORT_CLASS, e );
			}
			catch ( InstantiationException e ) {
				log.warn( "Could not instantiate JDBC support class " + JDBC4_SUPPORT_CLASS, e );
			}
		}
		else {
			log.info( "loading JDBC 3 support" );
			try {
				return ( JdbcSupport ) loadClass( JDBC3_SUPPORT_CLASS ).newInstance();
			}
			catch ( ClassNotFoundException e ) {
				log.warn( "Could not locate JDBC support class " + JDBC3_SUPPORT_CLASS, e );
			}
			catch ( IllegalAccessException e ) {
				log.warn( "Could not instantiate JDBC support class " + JDBC3_SUPPORT_CLASS, e );
			}
			catch ( InstantiationException e ) {
				log.warn( "Could not instantiate JDBC support class " + JDBC3_SUPPORT_CLASS, e );
			}
		}

		throw new IllegalStateException( "Could not determine proper JDBC support module" );
	}

	private static final Class[] NO_ARG_SIG = new Class[0];
	private static final Object[] NO_ARGS = new Object[0];

	private static boolean isJdbc4() {
		try {
			Class connectionClass = Connection.class;
			Method createClobMethod = connectionClass.getMethod( "createClob", NO_ARG_SIG );
			return createClobMethod.getDeclaringClass().equals( Connection.class );
		}
		catch ( NoSuchMethodException e ) {
			return false;
		}
	}

	private static boolean areLobCreationMethodsImplemented(Connection jdbcConnection) {
		try {
			Class connectionClass = Connection.class;
			Method createClobMethod = connectionClass.getMethod( "createClob", NO_ARG_SIG );
			Object clob = createClobMethod.invoke( jdbcConnection, NO_ARGS );
			try {
				Method freeMethod = clob.getClass().getMethod( "free", NO_ARG_SIG );
				freeMethod.invoke( clob, NO_ARGS );
			}
			catch ( Throwable ignore ) {
				log.trace( "Unable to free CLOB created to test createClob() implementation : " + ignore );
			}
			return true;
		}
		catch ( Throwable t ) {
			log.info( "createClob() method threw error : " + t );
			return false;
		}
	}

	private static Class loadClass(String className) throws ClassNotFoundException {
		return JdbcSupportLoader.class.getClassLoader().loadClass( className );
	}
}
