/*
 * Copyright (c) 2009 by Red Hat Inc and/or its affiliates or by
 * third-party contributors as indicated by either @author tags or express
 * copyright attribution statements applied by the authors.  All
 * third-party contributions are distributed under license by Red Hat Inc.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution; if not, write to:
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301  USA\
 */
package org.hibernate.engine.jdbc.jdbc4;

import java.sql.Blob;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Clob;
import java.sql.NClob;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;

import org.hibernate.engine.jdbc.LobCreationContext;
import org.hibernate.engine.jdbc.StreamUtils;

/**
 * TODO : javadoc
 *
 * @author Steve Ebersole
 */
public class ContextualLobCreator extends AbstractLobCreator {
	private LobCreationContext lobCreationContext;

	public ContextualLobCreator(LobCreationContext lobCreationContext) {
		this.lobCreationContext = lobCreationContext;
	}

	public Blob createBlob() {
		return ( Blob ) lobCreationContext.execute( BlobCreationCallback.INSTANCE );
	}

	public Blob createBlob(byte[] bytes) {
		try {
			Blob blob = createBlob();
			blob.setBytes( 1, bytes );
			return blob;
		}
		catch ( SQLException e ) {
			throw new IllegalStateException( "Unable to set BLOB bytes after creation", e );
		}
	}

	public Blob createBlob(InputStream inputStream, long length) {
		try {
			Blob blob = createBlob();
			OutputStream byteStream = blob.setBinaryStream( 1 );
			StreamUtils.copy( inputStream, byteStream );
			byteStream.flush();
			byteStream.close();
			// todo : validate length written versus length given?
			return blob;
		}
		catch ( SQLException e ) {
			throw new IllegalStateException( "Unable to prepare BLOB binary stream for writing", e );
		}
		catch ( IOException e ) {
			throw new IllegalStateException( "Unable to write stream contents to BLOB", e );
		}
	}

	public Clob createClob() {
		return ( Clob ) lobCreationContext.execute( ClobCreationCallback.INSTANCE );
	}

	public Clob createClob(String string) {
		try {
			Clob clob = createClob();
			clob.setString( 1, string );
			return clob;
		}
		catch ( SQLException e ) {
			throw new IllegalStateException( "Unable to set CLOB string after creation", e );
		}
	}

	public Clob createClob(Reader reader, int length) {
		try {
			Clob clob = createClob();
			Writer writer = clob.setCharacterStream( 1 );
			StreamUtils.copy( reader, writer );
			writer.flush();
			writer.close();
			return clob;
		}
		catch ( SQLException e ) {
			throw new IllegalStateException( "Unable to prepare CLOB stream for writing", e );
		}
		catch ( IOException e ) {
			throw new IllegalStateException( "Unable to write CLOB stream content", e );
		}
	}

	public NClob createNClob() {
		return ( NClob ) lobCreationContext.execute( NClobCreationCallback.INSTANCE );
	}

	public NClob createNClob(String string) {
		try {
			NClob clob = createNClob();
			clob.setString( 1, string );
			return clob;
		}
		catch ( SQLException e ) {
			throw new IllegalStateException( "Unable to set NCLOB string after creation", e );
		}
	}

	public NClob createNClob(Reader reader, int length) {
		try {
			NClob clob = createNClob();
			Writer writer = clob.setCharacterStream( 1 );
			StreamUtils.copy( reader, writer );
			writer.flush();
			writer.close();
			return clob;
		}
		catch ( SQLException e ) {
			throw new IllegalStateException( "Unable to prepare NCLOB stream for writing", e );
		}
		catch ( IOException e ) {
			throw new IllegalStateException( "Unable to write NCLOB stream content", e );
		}
	}

	private static class BlobCreationCallback implements LobCreationContext.Callback {
		private static BlobCreationCallback INSTANCE = new BlobCreationCallback();
		public Blob executeOnConnection(Connection connection) {
			try {
				return connection.createBlob();
			}
			catch ( SQLException e ) {
				throw new IllegalStateException( "Unable to create BLOB from JDBC connection", e );
			}
		}
	}

	private static class ClobCreationCallback implements LobCreationContext.Callback {
		private static ClobCreationCallback INSTANCE = new ClobCreationCallback();
		public Clob executeOnConnection(Connection connection) {
			try {
				return connection.createClob();
			}
			catch ( SQLException e ) {
				throw new IllegalStateException( "Unable to create CLOB from JDBC connection", e );
			}
		}
	}

	private static class NClobCreationCallback implements LobCreationContext.Callback {
		private static NClobCreationCallback INSTANCE = new NClobCreationCallback();
		public NClob executeOnConnection(Connection connection) {
			try {
				return connection.createNClob();
			}
			catch ( SQLException e ) {
				throw new IllegalStateException( "Unable to create NCLOB from JDBC connection", e );
			}
		}
	}
}
